/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
 Module:       JSBSimWorker.cs
 Author:       Lorne McIntosh
 Date started: 2007/07/31
 Purpose:      Wraps the functionality of JSBSimNET and makes it easy to thread

--------- Copyright (C) 2007  Lorne D. McIntosh (lorne@lorneswork.com) ---------

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

HISTORY
--------------------------------------------------------------------------------
2007/07/31   LDM   Created

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/

using System;
using System.ComponentModel;
using System.Diagnostics;
using System.Timers;

using JSBSimNET;

namespace JSBSimTest
{
    /// <summary>
    /// Wraps the functionality of JSBSim and makes it easy to thread
    /// </summary>
    public class JSBSimWorker : BackgroundWorker
    {
        // Define the root directory where the aircraft, engine, and scripts subdirectories are located
        private string rootDirectory;

        // The JSBSim FDM instance
        private FGFDMExecNET fdm;

        // A timer used for updating the GUI at set intervals
        private Timer updateStatusTimer;

        /// <summary>
        /// Contructor for JSBSimWorker
        /// </summary>
        /// <param name="rootDirectory">root directory where the aircraft, engine, and scripts subdirectories are located</param>
        public JSBSimWorker(string rootDirectory)
        {
            WorkerReportsProgress = false;
            WorkerSupportsCancellation = true;

            this.rootDirectory = rootDirectory;

            updateStatusTimer = new Timer();
            updateStatusTimer.Elapsed += delegate(object sender, ElapsedEventArgs e)
            {
                UpdateStatusEventHandler(sender, e, ref fdm);
            };
            updateStatusTimer.Interval = 250;
        }

        /// <summary>
        /// The "work" method for the worker thread.
        /// Calls LoadAndRunScript or LoadAndRunModel based on the number of arguments provided
        /// </summary>
        /// <param name="e">provide arguments in an Object[] array</param>
        protected override void OnDoWork(DoWorkEventArgs e)
        {
            Object[] args = (Object[])e.Argument;

            switch (args.Length)
            {
                case 1:
                    // 1 argument: assume user wants LoadAndRunScript
                    LoadAndRunScript(e, (string)args[0]);
                    break;

                case 2:
                    // 2 arguments: assume user wants LoadAndRunModel
                    LoadAndRunModel(e, (string)args[0], (string)args[1]);
                    break;

                default:
                    throw new Exception("provide 1 or 2 arguments in an object[] array");
            }
        }

        /// <summary>
        /// Loads and runs the model / reset file pair
        /// </summary>
        private void LoadAndRunModel(DoWorkEventArgs e, string modelFileName, string resetFileName)
        {
            fdm = new FGFDMExecNET();
            fdm.AircraftPath = rootDirectory + "\\aircraft";
            fdm.EnginePath = rootDirectory + "\\engine";

            fdm.LoadModel(modelFileName, true);
            fdm.LoadIC(resetFileName, true);
            fdm.DoTrim(1); //full trim

            bool result = fdm.RunIC();

            updateStatusTimer.Start();

            while (result && !(fdm.Holding() || fdm.IntegrationSuspended()))
            {
                // are we being asked to stop?
                if (CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                result = fdm.Run();
            }
            e.Result = "Simulation ended at time: " + fdm.SimTime;
        }

        /// <summary>
        /// Loads and runs the script file
        /// </summary>
        private void LoadAndRunScript(DoWorkEventArgs e, string scriptFileName)
        {
            string scriptPath = rootDirectory + "\\scripts\\";

            fdm = new FGFDMExecNET();
            fdm.AircraftPath = rootDirectory + "\\aircraft";
            fdm.EnginePath = rootDirectory + "\\engine";

            fdm.LoadScript(scriptPath + scriptFileName);

            bool result = fdm.RunIC();

            updateStatusTimer.Start();

            while (result && !(fdm.Holding() || fdm.IntegrationSuspended()))
            {
                // are we being asked to stop?
                if (CancellationPending)
                {
                    e.Cancel = true;
                    return;
                }

                result = fdm.Run();
            }
            e.Result = "Simulation ended at time: " + fdm.SimTime;
        }

        public static void UpdateStatusEventHandler( object sender, ElapsedEventArgs e, ref FGFDMExecNET fdm )
        {
            double alt = fdm.GetPropertyValue("position/h-sl-ft");
            Trace.WriteLine("=> Time: " + fdm.SimTime + " => Alt: " + alt);
        }

        protected override void OnRunWorkerCompleted(RunWorkerCompletedEventArgs e)
        {
            updateStatusTimer.Stop();
            base.OnRunWorkerCompleted(e);
        }
    }
}
